@@ -76,7 +76,7 @@ sub media_recv_echo {
 		}
 	    };
 
-	    $call->{loop}->addFD( $sock,
+	    $call->{loop}->addFD($sock, 0,
 		[ $echo_back,$s_sock,$addr,\@delay_buffer,$delay || 0,$writeto,{
 		    dtmf_gen => $args->{dtmf_events},
 		    dtmf_xtract => $mdtmf && $mdtmf->[$i] && $args->{cb_dtmf}
@@ -160,7 +160,7 @@ sub media_send_recv {
 		    CAN_NONBLOCKING or return;
 		}
 	    };
-	    $call->{loop}->addFD( $sock,
+	    $call->{loop}->addFD($sock, 0,
 		[
 		    $receive,
 		    $writeto,
